------------------------------------------
-- File:		Truck.lua
-- Description:
-- Date:		15/10/11 7:39 PM
--------------------------------------

-- Automatically included

-----------------------------------------------------------------------
engineSoundEnabled = true
-----------------------------------------------------------------------
function InitTruck(truck)
	-- common params
	truck.engineTension = 0
	truck.turbo = 0

	-- exhaust params
	truck.exhaustStartFactor = 0
	truck.exhaustCoolFactor = 0
	truck.exhaustLargeSmokeCooldown = 0
	truck.steamFactor = 0
	truck.isSteaming = false

	-- sound params
	truck.prevSteeringAngle = 0
	truck.prevPedalAccelerate = 0
    truck.prevPedalBrake = 0
	truck.isBreakPulled = false
	
	truck.prevGear = 0
	truck.wasHandbrake = false
	truck.wasDifferentialLocked = false	
	truck.wasAllWheelDrive = false	
	truck.wasReversing = false
	truck.engineState = "unknown"

	truck.linVelPrev = 0
	truck.prevEngineTension = 0
	truck.chassisPrevLinVel = m3MakeZero()
	truck.chassisPrevAngVel = m3MakeZero()	

	truck.wasEngineRev = false
	truck.elapsedGearTime = -1
	truck.elapsedStartTime = 0
	truck.elapsedDropletsTime = 0
end
-----------------------------------------------------------------------
function ProcessTruckEngineSound(truck, truckInput, elapsedTime)
	if (truckInput.isEngineIgnited) then
		if (truck.elapsedStartTime == 0) then
			C_SoundPlay(truck.soundEngineStart)
		end
		C_SoundStop(truck.soundEngineStop)		
		if (truck.soundEngineStartDelay == nil or truck.elapsedStartTime > truck.soundEngineStartDelay) then
			local engineTension = truck.engineTension
			if (not (truck.soundEngineStartDelay == nil)) then
				if (truck.elapsedStartTime > truck.soundEngineStartDelay + 1) then
					engineTension = math.max(0,
						engineTension, 1.0 - (math.max(truck.elapsedStartTime - (truck.soundEngineStartDelay + 0.5), 0) - 1.0)	/ 1.5, 0)		
				else
					engineTension = (truck.elapsedStartTime - truck.soundEngineStartDelay) / 1.0
				end
			end	
			C_ShakeTruck(truck.this, engineTension)
			-- debug state
			--C_RenderDebugString( string.format("state %s %.2f", truck.engineState, t) )
			--C_RenderDebugString( string.format("tension %.2f", engineTension) )			
			local newState = truck.engineState
			local tState = 0	
			
			local tLowMin = 0.02
			local tLowIn = 0.04
			local tLowOut = 0.06
			local tHightIn = 0.45
			local tHighOut = 0.35
			local tHighMax = 0.9
			if (truck.engineState == "idle") then
				if (engineTension > tLowIn) then
					newState = "low"
				end
			elseif (truck.engineState == "low") then
				if (engineTension > tHightIn) then
					newState = "high"
				elseif (truck.engineTension < tLowOut) then
					newState = "idle"
				else
					tState = saturate((engineTension - tLowMin) / (tHightIn - tLowMin))
				end
			elseif (truck.engineState == "high") then
				if (engineTension < tHighOut) then
					newState = "idle"
				else
					tState = saturate((engineTension - tHighOut) / (tHighMax - tHighOut))
				end
			else	
				newState = "idle"
			end
			-- debug state
			--C_WriteLog( truck.engineState )
			--C_RenderDebugString( string.format("state %s %.2f", truck.engineState, tState) )			
			if (not (newState == truck.engineState)) then
				if (truck.engineState == "idle") then
					C_SoundFadeOut(truck.soundEngineIdle, 1)
				elseif (truck.engineState == "low") then
					C_SoundFadeOut(truck.soundEngineLow, 1)
				elseif (truck.engineState == "high") then
					C_SoundFadeOut(truck.soundEngineHigh, 1)
					C_SoundFadeOut(truck.soundEngineHeavy, 1)
				end
				if (newState == "idle") then
					C_SoundPlay(truck.soundEngineIdle, true)
					C_SoundFadeIn(truck.soundEngineIdle, 0.1)
				elseif (newState == "low") then
					C_SoundPlay(truck.soundEngineLow, true)
					C_SoundFadeIn(truck.soundEngineLow, 1)
				elseif (newState == "high") then
					C_SoundPlay(truck.soundEngineHigh, true)
					C_SoundFadeIn(truck.soundEngineHigh, 1)

					C_SoundPlay(truck.soundEngineHeavy, true)
					C_SoundFadeIn(truck.soundEngineHeavy, 1)
				end
				truck.engineState = newState
			end
			if (truckInput.pedalAccelerate > 0.4 and truckInput.pedalAccelerate > truck.prevPedalAccelerate) then			
				if (not truck.wasEngineRev) then				
					C_SoundPlay(truck.soundEngineRev, false, true)			
					C_SoundSetVolumeMultiplier(truck.soundEngineRev, 0.8 + truckInput.pedalAccelerate * 0.8)
					truck.wasEngineRev = true
				end
			else			
				truck.wasEngineRev = false			
			end
			truck.prevPedalAccelerate = truckInput.pedalAccelerate

			if (truck.engineState == "low") then
				C_SoundSetFrequencyRatio(truck.soundEngineLow, 0.8 + tState * 0.4)
			elseif (truck.engineState == "high") then
				C_SoundSetFrequencyRatio(truck.soundEngineHigh, 0.8 + tState * 0.4)
				C_SoundSetVolumeMultiplier(truck.soundEngineHigh, 1 - saturate((tState - 0.6) / 0.8))				
				C_SoundSetVolumeMultiplier(truck.soundEngineHeavy, saturate((tState - 0.4) / 0.4))
			end		
			local turbo = saturate((engineTension - 0.01) * 0.8)
			if (turbo > truck.turbo) then
				truck.turbo = math.min(truck.turbo + elapsedTime * 0.4, turbo)
			else
				truck.turbo = math.max(truck.turbo - elapsedTime * 0.1, turbo)
			end

			if (truck.turbo > 0) then
				C_SoundPlay(truck.soundEngineTurbo, true)	
				
				C_SoundSetVolumeMultiplier(truck.soundEngineTurbo, saturate(truck.turbo * 0.5))
				C_SoundSetFrequencyRatio(truck.soundEngineTurbo, 0.2 + truck.turbo * 1.0)
			else
				C_SoundStop(truck.soundEngineTurbo)
			end	
			--C_RenderDebugString( string.format("tension %.2f turbo %.2f", engineTension, truck.turbo) )	
		--[[
			RenderDebugString( FormatString(L"Engine sound: Idle %d, Low %d, High %d, Heavy",
				m_pEngineIdle->IsPlaying(),
				m_pEngineLow->IsPlaying(),
				m_pEngineHigh->IsPlaying(),
				m_pEngineHeavy->IsPlaying()).c_str() )
		]]
			truck.prevEngineTension = engineTension

			if (truckInput.gear < 0 and engineTension > 0.05) then
				if (not truck.wasReversing) then
					C_SoundPlay(truck.soundReverse, true)
					C_SoundFadeIn(truck.soundReverse, 0.5)
				end
				C_SoundSetVolumeMultiplier(truck.soundReverse, 0.8 + engineTension * 0.4)
				C_SoundSetFrequencyRatio(truck.soundReverse, 0.8 + engineTension * 0.2)
				truck.wasReversing = true
			else
				if (truck.wasReversing) then
					C_SoundFadeOut(truck.soundReverse, 0.5)
				end
				truck.wasReversing = false
			end	
		end
	else
		C_ShakeTruck(truck.this, 0)
		if (truck.elapsedStartTime > 0) then
			C_SoundPlay(truck.soundEngineStop)
		end	
		C_SoundStop(truck.soundEngineIdle)
		C_SoundStop(truck.soundEngineLow)
		C_SoundStop(truck.soundEngineHigh)
		C_SoundStop(truck.soundEngineHeavy)
		C_SoundStop(truck.soundEngineTurbo)
		C_SoundStop(truck.soundEngineStart)
		C_SoundStop(truck.soundReverse)
	end
end
-----------------------------------------------------------------------
function ProcessTruckSound(truck, truckInput, wheels, elapsedTime)
	-- brake
	if (truckInput.pedalBrake > 0.05 and truckInput.pedalBrake > truck.prevPedalBrake) then
		if (not truck.isBreakPulled) then
			C_SoundPlay(truck.soundBrakePull)
			truck.isBreakPulled = true
		end
	end
	if (truckInput.pedalBrake < 0.10 and truckInput.pedalBrake < truck.prevPedalBrake) then
		if (truck.isBreakPulled) then
			C_SoundPlay(truck.soundBrakeRelease)
			C_SoundSetVolumeMultiplier(truck.soundBrakeRelease, 2.1)
			
			truck.isBreakPulled = false
		end
	end	
	truck.prevPedalBrake = truckInput.pedalBrake
	
	if (truckInput.isHandbrake and not truck.wasHandbrake) then
		C_SoundPlay(truck.soundHandbrake)
	end
	truck.wasHandbrake = truckInput.isHandbrake
		
	-- transmission
	if (truck.elapsedGearTime >= 0) then
		truck.elapsedGearTime = truck.elapsedGearTime + elapsedTime
	end
	if (not (truckInput.gear == truck.prevGear)) then
		C_SoundPlay(truck.soundGear)
		C_SoundSetVolumeMultiplier(truck.soundGear, 1.5)

		truck.elapsedGearTime = 0
	end
	truck.prevGear = truckInput.gear	
	if (truck.elapsedGearTime > 0.5) then
		if (math.random(10) <= 2) then
			C_SoundPlay(truck.soundCrackle)
		end		
		truck.elapsedGearTime = -1.0
	end
	
	if (not (truckInput.isDifferentialLocked == truck.wasDifferentialLocked)) then
		C_SoundPlay(truck.soundDiffLock)
		C_ShakeDriveCamera(truck.this, 0.5, 0.5)
	end
	truck.wasDifferentialLocked = truckInput.isDifferentialLocked
	if (not (truckInput.isAllWheelDrive == truck.wasAllWheelDrive)) then
		C_SoundPlay(truck.soundDiffLock)		
		C_ShakeDriveCamera(truck.this, 0.5, 0.5)
	end
	truck.wasAllWheelDrive = truckInput.isAllWheelDrive
	if (engineSoundEnabled) then
		ProcessTruckEngineSound(truck, truckInput, elapsedTime)
	end

	local linVelL = m3Length(truckInput.linVelVector)
	if (linVelL > 1.0) then
		local vA = m3Diff(truckInput.linVelVector, truck.chassisPrevLinVel)
		m3Div(vA, elapsedTime)

		local fA = m3Dot(truckInput.linVelVector, vA) / linVelL
		if (fA < -2.0) then
			C_SoundPlay(truck.soundAbruptStop)
			C_SoundSetVolumeMultiplier(truck.soundAbruptStop, clamp(-fA * 0.15, 0.6, 1.0) )
		end
	end
	local angVelL = m3Length(truckInput.angVelVector)
	if (angVelL > 0.5) then
		local vA = m3Diff(truckInput.angVelVector, truck.chassisPrevAngVel)
		m3Div(vA, elapsedTime)

		local fA = m3Length(vA)
		if (fA > 0.8) then
			--C_TraceDebugString( string.format("fA = %.2f", fA) )
			C_SoundPlay(truck.soundAbruptStop)
			C_SoundSetVolumeMultiplier(truck.soundAbruptStop, clamp(fA * 0.4, 0.6, 1.4) )
		end
	end	
	truck.chassisPrevAngVel = truckInput.angVelVector	
	truck.chassisPrevLinVel = truckInput.linVelVector

    -- play suspension stress sounds
	if (truck.suspStressFactors == nil) then		
		truck.wheelsLinVelPrev = {}
		truck.suspStressFactors = {}
		truck.waterDepthPrev = {}
		for i,wheel in ipairs(wheels) do
			truck.wheelsLinVelPrev[i] = wheel.linVel
			truck.suspStressFactors[i] = wheel.suspStressFactor
			truck.waterDepthPrev[i] = wheel.waterDepth
		end
	end	
	
	local wheelsDirt = 0
	local wheelsGrass = 0
	local wheelsMud = 0
	--local wheelsSkid = 0
	local wheelsSpinning = 0
	local wheelsSteer = 0
	local wheelsWater = 0
	local waterHit = 0
	
	for i,wheel in ipairs(wheels) do
		local dSusp = math.abs(wheel.suspStressFactor - truck.suspStressFactors[i]) / elapsedTime
		if (dSusp > 1) then
			--C_TraceDebugString( "STRESS" )
			C_SoundPlay(truck.soundChassisStress)
			C_SoundSetVolumeMultiplier(truck.soundChassisStress, clamp(dSusp * 0.1, 0.4, 1.0))
		end
		truck.suspStressFactors[i] = wheel.suspStressFactor
		
		local angVelParam = math.abs(wheel.angVelDamped) / 8
		if (debrisType == "fragments_grass") then
			wheelsGrass = math.max(wheelsGrass, angVelParam * wheel.contactFriction)
		else
			wheelsDirt = math.max(wheelsDirt, angVelParam * wheel.contactFriction)
		end
		if (wheel.contactFriction < EPSILON) then
			wheelsSpinning = math.max(wheelsSpinning, angVelParam * 0.5)
		end		
		wheelsMud = math.max(wheelsMud, angVelParam * math.min(wheel.mudDepth / 0.5, 1))
		wheelsWater = math.max(wheelsWater, angVelParam * math.min(wheel.waterDepth / 0.5, 1))	
		wheelsSteer = math.max(wheelsSteer, wheel.contactFriction * 0.5 * saturate(1.0 - wheel.waterDepth / 0.2))
	
		local dLinVel = math.abs(wheel.linVel - truck.wheelsLinVelPrev[i]) / elapsedTime
		--if (dLinVel > 8) then			
		--	wheelsSkid = wheelsSkid + dLinVel * wheel.contactFriction * 0.01
		--end
		truck.wheelsLinVelPrev[i] = wheel.linVel

		local dWater = math.abs(wheel.waterDepth - truck.waterDepthPrev[i]) / elapsedTime
		waterHit = math.max(waterHit, dLinVel * 0.02 * wheel.waterDepth + dWater * 0.2)
		
		truck.waterDepthPrev[i] = wheel.waterDepth
	end	
	wheelsDirt = saturate(wheelsDirt - 0.1)
	wheelsGrass = saturate(wheelsGrass - 0.1)
	wheelsMud = saturate(wheelsMud - 0.1)
	wheelsWater = saturate(wheelsWater - 0.1)
	waterHit = saturate(waterHit - 0.1)
	
	if (wheelsDirt > 0) then
		C_SoundPlay(truck.soundWheelDirt, true)
		C_SoundFadeIn(truck.soundWheelDirt, 0.1)
		C_SoundSetVolumeMultiplier(truck.soundWheelDirt, 0.4 + wheelsDirt * 0.8)	
	else
		C_SoundFadeOut(truck.soundWheelDirt, 0.1)
	end
	if (wheelsGrass > 0) then
		C_SoundPlay(truck.soundWheelGrass, true)
		C_SoundFadeIn(truck.soundWheelGrass, 0.1)
		C_SoundSetVolumeMultiplier(truck.soundWheelGrass, 0.4 + wheelsGrass * 0.8)	
	else
		C_SoundFadeOut(truck.soundWheelGrass, 0.1)
	end
	--[[if (wheelsSkid > 0) then
		C_SoundPlay(truck.soundWheelSkid)
		C_SoundSetVolumeMultiplier(truck.soundWheelSkid, 0.4 + wheelsSkid * 0.8)	
	end	]]
	if (wheelsSpinning > 0.1) then
		C_SoundPlay(truck.soundWheelSpinning, true)
		C_SoundFadeIn(truck.soundWheelSpinning, 0.1)
		C_SoundSetVolumeMultiplier(truck.soundWheelSpinning, wheelsSpinning)	
	else
		C_SoundFadeOut(truck.soundWheelSpinning, 0.1)
	end	
	if (wheelsMud > 0 and wheelsMud < 0.5) then		
		C_SoundPlay(truck.soundWheelMud, true)
		C_SoundFadeIn(truck.soundWheelMud, 0.1)
		C_SoundSetVolumeMultiplier(truck.soundWheelMud, 0.4 + wheelsMud * 1.2)	
	else
		C_SoundFadeOut(truck.soundWheelMud, 0.1)
	end		
	if (wheelsMud > 0.5) then
		C_SoundPlay(truck.soundWheelExtrude, true)
		C_SoundFadeIn(truck.soundWheelExtrude, 0.1)
		C_SoundSetVolumeMultiplier(truck.soundWheelExtrude, 0.4 + (wheelsMud - 0.5) * 2.0)	
	else
		C_SoundFadeOut(truck.soundWheelExtrude, 0.1)
	end	
	if (wheelsWater > 0) then
		C_SoundPlay(truck.soundWheelWater, true)
		C_SoundFadeIn(truck.soundWheelWater, 0.2)
		C_SoundSetVolumeMultiplier(truck.soundWheelWater, 0.4 + wheelsWater * 0.6)	
	else
		C_SoundFadeOut(truck.soundWheelWater, 0.05)
	end		
	if (waterHit > 0 and not C_IsSoundPlaying(truck.soundWaterHit)) then
		--C_TraceDebugString( "HIT" )
		C_SoundPlay(truck.soundWaterHit)
		C_SoundSetVolumeMultiplier(truck.soundWaterHit, 0.6 + waterHit * 0.8)	
	end	
	
	-- wheel steering noise
	local deltaSteer = wheelsSteer * math.abs(truckInput.steeringAngle - truck.prevSteeringAngle) / elapsedTime
	--C_RenderDebugString( string.format("%.2f", deltaSteer) )
	if (deltaSteer > 0.05) then
		C_SoundPlay(truck.soundWheelSteer, true)
		C_SoundFadeIn(truck.soundWheelSteer, 0.1)
		C_SoundSetVolumeMultiplier(truck.soundWheelSteer, math.min((deltaSteer - 0.05) / 0.6, 1.0))
	else
		C_SoundFadeOut(truck.soundWheelSteer, 0.1)
	end
	truck.prevSteeringAngle = truckInput.steeringAngle
end
-----------------------------------------------------------------------
function ProcessTruckExhaust(truck, truckInput, wheels, exhausts, steams, elapsedTime)
	local linVelL = m3Length(truckInput.linVelVector)
	local isAnySubmerged = false
	if (table.getn(exhausts) > 0) then
		local exhaustAmount = saturate(truck.engineTension * 8.0 - 1.0)		
		local engineStartExhaust = 0	

		if (not (truck.soundExhaustStartTime == nil) and not (truck.soundExhaustFadeStartTime == nil) and not (truck.soundExhaustStopTime == nil) and truckInput.isEngineIgnited) then
			if (truck.elapsedStartTime < truck.soundExhaustFadeStartTime) then
				if (truck.elapsedStartTime > truck.soundExhaustStartTime) then
					engineStartExhaust = 1.0
				end
			elseif (truck.soundExhaustFadeStartTime < truck.soundExhaustStopTime) then
				engineStartExhaust = 1.0 - math.min((truck.elapsedStartTime - truck.soundExhaustFadeStartTime) / (truck.soundExhaustStopTime - truck.soundExhaustFadeStartTime), 1.0)
			end
		end
		exhaustAmount = math.min(exhaustAmount + engineStartExhaust, 1.0)		
		
		local exhaustThreshold = 0.06
		if (exhaustAmount > exhaustThreshold) then
			local d = (exhaustAmount - exhaustThreshold) / (1.0 - exhaustThreshold) * elapsedTime
			truck.exhaustStartFactor = math.min(truck.exhaustStartFactor + d * 0.8, 2.0)
			truck.exhaustCoolFactor = math.min(truck.exhaustCoolFactor + d * 0.8, 16.0)
		else
			local d = (exhaustThreshold - exhaustAmount) / exhaustThreshold * elapsedTime
			truck.exhaustStartFactor = math.max(truck.exhaustStartFactor - d * 2.0, 0)
			truck.exhaustCoolFactor = math.max(truck.exhaustCoolFactor - d * 0.6, 0)
		end
		--C_RenderDebugString( string.format("exhaustStartFactor %.2f exhaustCoolFactor %.2f", truck.exhaustStartFactor, truck.exhaustCoolFactor) )
		
		local exhaustLargeSmokeCooldownTime = 0.6
		if (exhaustAmount > 0.25 and truck.exhaustCoolFactor < 8.0) then
			truck.exhaustLargeSmokeCooldown = math.min(truck.exhaustLargeSmokeCooldown + elapsedTime, exhaustLargeSmokeCooldownTime)
		else
			truck.exhaustLargeSmokeCooldown = math.max(truck.exhaustLargeSmokeCooldown - elapsedTime, 0)
		end
		
		local tStart = math.max(0, 1.0 - truck.exhaustStartFactor)
		local velocityMultiplier = (0.8 + exhaustAmount * 0.2) * (0.8 + tStart * 0.2)
		-- More exhaust if there were none for a short period of time
		exhaustAmount = exhaustAmount * (0.8 + tStart * 0.6)
		--exhaustAmount = 1	
		--C_RenderDebugString( string.format("exhaustAmount %.2f", exhaustAmount) )

		-- larger particles when truck is moving fast
		local exhaustScale = 1.0 + clamp((linVelL - 4.0) / 8.0, 0.0, 2.0)	
					
		for i,exhaust in ipairs(exhausts) do
			local exhaustSubmergePoint = m3CopyVec(exhaust.exhaustOrg)
			exhaustSubmergePoint.y = exhaustSubmergePoint.y - 0.2		
			
			local isSubmerged = C_IsSubmergedAt(exhaustSubmergePoint)
			if (isSubmerged) then	
				isAnySubmerged = true
			else
				local partVel = m3CopyVec(exhaust.exhaustVel)
				-- add chassis velocity before multiply
				m3Add(partVel, truckInput.linVelVector)
				m3Mul(partVel, velocityMultiplier)
				
				if (exhaustAmount > EPSILON) then			
					
					-- approximate area of dense smoke
					local bigExhaustOrg = m3MulMake(truckInput.linVelVector, -0.6)
					m3Add(bigExhaustOrg, exhaust.exhaustOrg)
					m3Add(bigExhaustOrg, m3MulMake(partVel, 0.6))
					
					--C_RenderDebugPoint(exhaust.exhaustOrg)				
					local numParticles = GetNumEmitParticles(0.04, exhaustAmount, elapsedTime)
					
					
					if (numParticles > 0) then
						if (table.getn(exhausts) == 3 and i == 3) then
							--nop 3d exh
						elseif (table.getn(exhausts) == 3) then
							C_SetParticlesBrand("stars")
							C_AddParticle(exhaust.exhaustOrg, partVel)
						else
							local exhaustOpacity = 0.2 + saturate(1 - (truck.exhaustCoolFactor - 2.0) / 4.0) * 0.4
							exhaustOpacity = exhaustOpacity + (1.0 - saturate(truck.exhaustCoolFactor / 0.5)) * 0.8
							exhaustOpacity = exhaustOpacity * 1.4
						
							-- Black exhaust when there were none for a long period of time
							local exhaustBrightness = math.max(0.1, 0.8 - tStart * tStart * 1.6)
							exhaustBrightness = saturate(exhaustBrightness - saturate((truckInput.damageFactor - 0.8) / 0.1))
						
							C_SetParticlesBrand("exhaust_small")
							for n = 1, numParticles  do
								C_AddParticle(exhaust.exhaustOrg, partVel, exhaustScale, exhaustOpacity, exhaustBrightness)
							end
							if (truck.exhaustLargeSmokeCooldown == exhaustLargeSmokeCooldownTime) then				
								C_SetParticlesBrand("exhaust_large")						
								C_AddParticle(bigExhaustOrg, m3MulMake(partVel, 0.25), exhaustScale, exhaustOpacity, exhaustBrightness)							
							end
						end
					end
				end
				
				--local numHeatParticles = GetNumEmitParticles(0.1, saturate(truck.engineTension * 8.0 - 1.0), elapsedTime)
				if (truck.elapsedStartTime > truck.soundEngineStartDelay and not table.getn(exhausts) == 3) then
					local numHeatParticles = GetNumEmitParticles(0.1, 1, elapsedTime)
					if (numHeatParticles > 0) then	
					--if (math.random() < 0.1) then
						local exhaustOpacity = 0.5 + saturate(truck.engineTension * 8.0) * 0.5
						C_SetParticlesBrand("exhaust_heat")						
						C_AddParticle(exhaust.exhaustOrg, partVel, exhaustScale, exhaustOpacity)						
					end		
				end
			end
		end
		if (truck.exhaustLargeSmokeCooldown == exhaustLargeSmokeCooldownTime) then	
			truck.exhaustLargeSmokeCooldown = 0
		end	
	end
	if (truckInput.isEngineIgnited and isAnySubmerged) then
		C_SoundPlay(truck.soundWaterGurgle, true)
		C_SoundFadeIn(truck.soundWaterGurgle, 0.2)
	else
		C_SoundFadeOut(truck.soundWaterGurgle, 0.2)
	end
	
	-- engine and wheels steam
	for i,steam in ipairs(steams) do
		local steamDamage = 0
		if (truckInput.isEngineIgnited) then
			steamDamage = saturate((truckInput.damageFactor - 0.8) / 0.2)
		end	
		if (math.random() < saturate((truckInput.damageFactor - 0.8) / 0.2) * truckInput.torqueFactor * 0.05) then
			local partFireOrg = m3CopyVec(steam.steamOrg)
			partFireOrg.x = partFireOrg.x + (math.random() - 0.5) * 0.5
			partFireOrg.y = partFireOrg.y + (math.random() - 0.5) * 0.5
			partFireOrg.z = partFireOrg.z + (math.random() - 0.5) * 0.5
						
			local partFireVel = m3MulMake(truckInput.linVelVector, 0.8)			
			local partSparksVel = m3MakeVec( 
						(math.random() - 0.5) * 5,
						0.5 + math.random() * 3,
						(math.random() - 0.5) * 5)
			m3Mul(partSparksVel, steam.steamScale)
			m3Add(partSparksVel, partFireVel)	
			partFireVel.y = partFireVel.y + 1			
			--partFireVel.x = partFireVel.x + (math.random() - 0.5) * 0.2
			--partFireVel.z = partFireVel.z + (math.random() - 0.5) * 0.2
			
			local numSparksParticle = GetNumEmitParticles(0.001, steamDamage * steam.steamScale, elapsedTime)
			if (numSparksParticle > 0) then	
				C_SetParticlesBrand("engine_sparks")
				for n = 1, numSparksParticle  do
					local partVel = m3CopyVec(partSparksVel) 
					partVel.x = partVel.x + (math.random() - 0.5) * 2
					partVel.y = -math.random() * 2
					partVel.z = partVel.z + (math.random() - 0.5) * 2						
					C_AddParticle(partFireOrg, partVel, 1.0 + truck.steamFactor * 0.2, 0.4 + steamDamage * 0.6)
				end	
			end
			partFireOrg.y = partFireOrg.y - 0.5
			
			C_SetParticlesBrand("engine_fire")				
			C_AddParticle(partFireOrg, partFireVel, 1.0 + truck.steamFactor * 0.2, 0.4 + steamDamage * 0.6)		
			C_SoundPlay(truck.soundDamageCrackle)			
		end
					
		--steamDamage = 1
		local steamAmount = steamDamage * (0.5 + truckInput.torqueFactor)
		if (truck.isSteaming) then
			steamAmount = math.max(steamAmount, truck.steamFactor)
			
			truck.steamFactor = math.max(truck.steamFactor - 0.06 * elapsedTime, 0)
			if (truck.steamFactor == 0) then
				truck.isSteaming = false
			end
		else
			local steamSubmergePoint = m3CopyVec(steam.steamOrg)
			steamSubmergePoint.y = steamSubmergePoint.y - 0.2
			
			truck.isSteaming = C_IsSubmergedAt(steamSubmergePoint)
			truck.steamFactor = saturate(truck.steamFactor + (truck.engineTension * 0.1 - 0.01) * elapsedTime)	
		end
		if (steamAmount > 0) then
			C_SetParticlesBrand("steam")
			
			local numParticles = GetNumEmitParticles(0.14, steamAmount * steam.steamScale, elapsedTime)
			for n = 1, numParticles  do
				local partVel = m3MakeVec( 
					(math.random() - 0.5) * 1.5,
					0.5 + math.random() * 1,
					(math.random() - 0.5) * 1.5)
				m3Mul(partVel, steam.steamScale)
				m3Add(partVel, m3MulMake(truckInput.linVelVector, 0.8))	
				
				local partOrg = m3CopyVec(steam.steamOrg)
				partOrg.x = partOrg.x + (math.random() - 0.5)
				partOrg.y = partOrg.y + (math.random() - 0.8)
				partOrg.z = partOrg.z + (math.random() - 0.5)
				C_AddParticle(partOrg, partVel, 1.0 + truck.steamFactor * 0.2, 0.4 + steamAmount * 0.6)
			end	
		end
		local heatFactor = saturate((truck.steamFactor - 0.6) / 0.2)
		if (linVelL < 0.2 and heatFactor > 0) then			
			local numHeatParticles = GetNumEmitParticles(0.3, heatFactor, elapsedTime)
			if (numHeatParticles > 0) then	
				C_SetParticlesBrand("steam_heat")
				for n = 1, numHeatParticles  do
					local partVel = m3MakeVec( 
						(math.random() - 0.5) * 1.5,
						0.5 + math.random() * 1,
						(math.random() - 0.5) * 1.5)
					m3Mul(partVel, steam.steamScale)
					m3Add(partVel, m3MulMake(truckInput.linVelVector, 0.8))	
					
					local partOrg = m3CopyVec(steam.steamOrg)
					partOrg.x = partOrg.x + (math.random() - 0.5)
					partOrg.y = partOrg.y + (math.random() - 0.8)
					partOrg.z = partOrg.z + (math.random() - 0.5)
					C_AddParticle(partOrg, partVel, 1.0, 0.4 + heatFactor * 0.6)
				end	
			end		
		end
		--C_RenderDebugString( string.format("steamFactor %.2f steamDamage %.2f", truck.steamFactor, steamDamage) )
	end	
	
	if (truck.wheelsSteamFactors == nil) then
		truck.wheelsSteamFactors = {}
		truck.wheelsHeatFactors = {}
		truck.wheelsSteaming = {}
		for i,wheel in ipairs(wheels) do
			truck.wheelsSteamFactors[i] = 0
			truck.wheelsHeatFactors[i] = 0
			truck.wheelsSteaming[i] = false
		end
	end	
	
	for iWheel,wheel in ipairs(wheels) do	
		local velSpin = math.abs(wheel.angVelDamped) * wheel.radius - math.abs(wheel.linVel)
		if (velSpin > 0.1) then
			truck.wheelsHeatFactors[iWheel] = math.min(truck.wheelsHeatFactors[iWheel] + 
				wheel.contactFriction * velSpin / 12 * 0.01, 2)
		else 
			truck.wheelsHeatFactors[iWheel] = math.max(truck.wheelsHeatFactors[iWheel] - 0.004, 0)		
		end
		local heatSteam = saturate((truck.wheelsHeatFactors[iWheel] - 0.5) / 1.5)
		if (truck.wheelsSteaming[iWheel] or heatSteam > 0) then
			--truck.wheelsSteamFactors[iWheel] = 1			
			local steamAmount = math.max(truck.wheelsSteamFactors[iWheel], heatSteam)
			if (steamAmount > 0) then
				local numParticles = GetNumEmitParticles(0.2, steamAmount, elapsedTime)
				if (numParticles > 0) then
					local partScale = wheel.radius / 0.6
					for n = 1, numParticles  do		
						local partOrg = m3CopyVec(wheel.axleDir)
						if (wheel.isRightSided) then
							m3Negate(partOrg)
						end	
						m3Mul(partOrg, wheel.width * 0.5)
						m3Add(partOrg, wheel.org)
						partOrg.y = partOrg.y + wheel.radius * 0.5
						
						local partVel = m3MakeVec( 
							(math.random() - 0.5) * 0.6,
							0.2 + math.random() * 0.4,
							(math.random() - 0.5) * 0.6)
						m3Add(partVel, m3MulMake(truckInput.linVelVector, 0.8))	
						C_AddParticle(partOrg, partVel, partScale, 0.2 + steamAmount * 0.4)
					end
				end
				truck.wheelsSteamFactors[iWheel] = math.max(truck.wheelsSteamFactors[iWheel] - 0.06 * elapsedTime, 0)			
			else
				truck.wheelsSteaming[iWheel] = false
			end
		else
			truck.wheelsSteaming[iWheel] = C_IsSubmergedAt(wheel.org)
			truck.wheelsSteamFactors[iWheel] = saturate(truck.wheelsSteamFactors[iWheel] + 
				(truckInput.pedalBrake * linVelL * 0.4 + math.abs(wheel.angVelDamped) * 0.008 - 0.01) * elapsedTime)	
		end
		--C_RenderDebugString( string.format("wheelsSteamFactors %.2f wheelsHeatFactors %.2f", truck.wheelsSteamFactors[iWheel], truck.wheelsHeatFactors[iWheel]) )
	end		
end
-----------------------------------------------------------------------
function SpawnWheelsParticles(truck, linVelVector, wheels, elapsedTime)
    local numWheelsFactor = 1
    if (table.getn(wheels) > 6) then
        numWheelsFactor = math.max(1 - (table.getn(wheels) - 6) / 4, 0.5)
    end	
    for iWheel,wheel in ipairs(wheels) do		
        if (wheel.waterDepth < 0.1) then
			-- Dust particles
			if (wheel.dustCoef - wheel.wetnessCoef > EPSILON) then
				local dustAmount = math.max( math.abs(wheel.angVel) * wheel.radius, wheel.linVel ) - 0.5
				dustAmount = saturate(dustAmount * 0.1) * wheel.dustCoef * (1.0 - wheel.mudCoef)
				-- fewer dust when position remains the same
				dustAmount = dustAmount * saturate(wheel.linVel / 2.0 + 0.2) * wheel.contactFriction
				
				local numDustParticles = GetNumEmitParticles(0.06, dustAmount * numWheelsFactor, elapsedTime)
				--C_RenderDebugString( string.format("dustAmount %.2f", dustAmount) )

				local dustOrg = m3MakeVec(
					wheel.org.x,
					wheel.org.y - wheel.radius,
					wheel.org.z)
				local dustOpacity = 0.2 + dustAmount * 0.4
				local dustScale = 0.5 + dustAmount * 1.5
				--C_RenderDebugString( string.format("dustAmount %.2f", dustAmount) )
				
				C_SetParticlesBrand("dust_small")
				for n = 1, numDustParticles  do
					C_AddParticle(dustOrg, m3MakeVec(0,1,0), dustScale, dustOpacity)
				end
			end

			-- Debris particles
			local debrisAmount = 0
			if (wheel.mudCoef > 0.25 or not(wheel.debrisType == "")) then
				local angLinVel = math.abs(wheel.angVel) * wheel.radius
				if (wheel.debrisType == "fragments_grass") then
					debrisAmount = saturate((wheel.contactFriction * angLinVel - 1) * 0.2)			
				else
					debrisAmount = saturate((wheel.contactFriction * angLinVel - 5.0) * 0.5) * 0.5
					-- debris from mud
					debrisAmount = math.max(debrisAmount, saturate((wheel.mudCoef * angLinVel - 2.0) * 0.2))
				end
				--debrisAmount = 1
				--C_RenderDebugString( string.format("debrisAmount %.2f", debrisAmount) )
				if (wheel.mudCoef > 0.25) then	
					C_SetParticlesBrand("fragments_dirt")
				else   
					C_SetParticlesBrand(wheel.debrisType)
				end				
				if (debrisAmount > EPSILON) then	
					local numDebrisParticles = GetNumEmitParticles(0.03, debrisAmount * numWheelsFactor, elapsedTime)
					local mudDepthT = math.min(wheel.mudDepth / wheel.radius * 1.5, 1.0)
					--C_RenderDebugString( string.format("mudDepthT %.2f", mudDepthT) )
					local emitVel = math.min(mudDepthT * 2 + angLinVel, 4) * 0.6
				
					local basisCos = wheel.wheelDir
					local basisSin = wheel.suspDir
					for n = 1, numDebrisParticles do
						local a = 0.5 * (mudDepthT - 1 + math.random()) * math.pi
						if (wheel.angVel > 0) then
							a = math.pi - a
						end
						local r = m2MakeVec(math.cos(a), math.sin(a))
						
						local partOrg = m3CopyVec(wheel.axleDir)
						m3Mul(partOrg, (math.random() - 0.5) * wheel.width)

						local rOffset = m3MulMake(basisCos, r.x)
						m3Add(rOffset, m3MulMake(basisSin, r.y))

						m3Add(partOrg, m3MulMake(rOffset, wheel.radius))
						m3Add(partOrg, wheel.org)

						-- perpendicular radial coords
						local r_p = m2MakePerpCW(r)
						local partVel = m3MulMake(basisCos, r_p.x)
						m3Add(partVel, m3MulMake(basisSin, r_p.y))
						if (wheel.angVel < 0) then
							m3Negate(partVel)
						end
						
						m3Add(partVel, m3MulMake(rOffset, 0.2))
						m3Add(partVel, m3MulMake(wheel.axleDir, 0.2 * (math.random() - 0.5)))
						m3Mul(partVel, emitVel + math.random() * 0.2)               
						m3Add(partVel, m3MulMake(linVelVector, 0.6))
				   
						C_AddParticle(partOrg, partVel)					
					end
				end	

				if (not (wheel.debrisType == "")) then				
					-- sliding dirt particles
					local axleVel = m3Dot(wheel.linVelVector, wheel.axleDir)
					local slidingAmount = saturate((math.abs(axleVel) - 1.5) * 0.3) * wheel.contactFriction				
					--C_RenderDebugString( string.format("wheel %d - slidingAmount %.2f", iWheel - 1, slidingAmount) )	
					
					if (slidingAmount > 0) then
						local emitVel = m3MulMake(wheel.axleDir, 0.4)	
						if (axleVel < 0) then
							m3Negate(emitVel)
						end
						m3Add(emitVel, linVelVector)
						
						local partDir = m3CopyVec(wheel.axleDir)	
						if (axleVel < 0) then
							m3Negate(partDir)
						end
						m3Add(partDir, m3MulMake(wheel.suspDir, 0.5 + slidingAmount))
						m3Normalize(partDir)
						
						local numBurstParticles = GetNumEmitParticles(0.015, slidingAmount, elapsedTime)	
						C_SetParticlesBrand("dirt_burst_medium")		
						for n = 1, numBurstParticles do
							local partOrg = m3CopyVec(wheel.org)
							partOrg.y = partOrg.y - wheel.radius

							local partVel = m3CopyVec(emitVel)										
							partVel.x = partVel.x + (math.random() - 0.5) * 2
							partVel.y = partVel.y + math.random() * 2.0 + slidingAmount * 2
							partVel.z = partVel.z + (math.random() - 0.5) * 2
							
							local partScale = 0.6 + math.random() * 1.2
							-- ural as a reference
							partScale = partScale * wheel.radius / 0.58
							
							C_AddParticleDir(partOrg, partVel, partDir, partScale)						
						end			
					end						
				end
			end			
		end	
    end
end
-----------------------------------------------------------------------
function SpawnWaveParticles(truck, truckInput, wave, elapsedTime)
	local linVelL = m3Length(truckInput.linVelVector)
	local waveOffset = m3Diff(wave.v1, wave.v0)
	local waveLength = m3Length(waveOffset)
	local emitBurstAmount =	
		saturate((math.abs(linVelL) - 2.0) / 2.0) * 
		saturate(waveLength * 0.5)			
	local numBurstParticles = GetNumEmitParticles(0.02, emitBurstAmount, elapsedTime)
	
	if (numBurstParticles >= 1) then
		for n = 1, numBurstParticles  do
			local partScale = 0.5 + saturate((math.abs(linVelL) - 1.0) / 8.0) * 0.5
			local t = math.random()
			local partOrg = m3MulMake(waveOffset, t)
			m3Add(partOrg, wave.v0)
			partOrg.y = partOrg.y - 0.25			
			
			--[[if (truckInput.linVelL > 3.0) then
				local partVel = m3MulMake(truckInput.linVelVector, 1.2)	
				partVel.x = partVel.x + (math.random() - 0.5)
				partVel.y = emitBurstAmount * 4.0 + 2.0
				partVel.z = partVel.z + (math.random() - 0.5)
				if (partVel.y > 4.0) then
					partVel.y = 4.0
				end
				if (t > 0.5) then
					m3Add(partVel, m3MulMake(waveOffset, (1.5 - t) * 2.0 / waveLength))
				else
					m3Add(partVel, m3MulMake(waveOffset, -(0.5 + t) * 2.0 / waveLength))
				end
				if (math.random() < 0.5) then
					C_SetParticlesBrand("water_burst_large")
				else
					C_SetParticlesBrand("water_burst_medium")
				end
				C_AddParticleColor(partOrg, partVel, truckInput.waterColor)			
			end]]
			local partFluidVel = m3MulMake(truckInput.linVelVector, 1.4)	
			partFluidVel.x = partFluidVel.x + (math.random() - 0.5) * 2
			partFluidVel.y = emitBurstAmount * 0.5 + 3.0
			partFluidVel.z = partFluidVel.z + (math.random() - 0.5) * 2
			
			C_SetParticlesBrand("water_fluid_large")
			C_AddParticleColor(partOrg, partFluidVel, truckInput.waterColor)			
		end
	end	
	if (emitBurstAmount > 0.1) then
		C_SoundPlay(truck.soundWaterSplashes, true)
		C_SoundFadeIn(truck.soundWaterSplashes, 0.2)	
		C_SoundSetVolumeMultiplier(truck.soundWaterSplashes, 0.4 + emitBurstAmount * 0.8)
	else
		C_SoundFadeOut(truck.soundWaterSplashes, 0.2)	
	end

	if (not(wave.flow0 == nil) and not(wave.flow1 == nil)) then		
		local flowOffset = m3Diff(wave.flow1, wave.flow0)
		local flowLength = m3Length(flowOffset)
		local d = math.min(0.1 / flowLength, 0.1)

		local numFlowParticles = GetNumEmitParticles(0.05, flowLength * wave.flowIntensity, elapsedTime)
		if (numFlowParticles >= 1) then	
			--local partVel = m3CopyVec(truckInput.linVelVector)
			local partVel = m3MulMake(truckInput.linVelVector, 0.4)	
			partVel.y = partVel.y - 1.5
				
			C_SetParticlesBrand("water_flow")
			for n = 1, numFlowParticles  do
				--C_TraceDebugString( "FLOW" )
				local t = math.random()				
				local partOrg = m3MulMake(flowOffset, t * (1 - d) + d)
				m3Add(partOrg, wave.flow0)				
				C_AddParticleColor(partOrg, partVel, truckInput.waterColor, 1, wave.flowIntensity)				
			end
		end	
		C_SoundPlay(truck.soundWaterDropping, true)
		C_SoundFadeIn(truck.soundWaterDropping, 0.2)		
	else
		C_SoundFadeOut(truck.soundWaterDropping, 0.2)
	end	
end
-----------------------------------------------------------------------
function ProcessTruck(truck, truckInput, wheels, exhausts, steams, wave, elapsedTime)
	if (elapsedTime > EPSILON) then
		--[[if (truckInput.torqueFactor > truck.engineTension)then
			truck.engineTension = lerp(truck.engineTension, truckInput.torqueFactor, math.min(1.0, elapsedTime * 60.0 * 0.2))
		else
			truck.engineTension = lerp(truck.engineTension, truckInput.torqueFactor, math.min(1.0, elapsedTime * 60.0 * 0.1))
		end]]
		truck.engineTension = lerp(truck.engineTension, truckInput.torqueFactor, math.min(1.0, elapsedTime * 5.0))
		
		ProcessTruckSound(truck, truckInput, wheels, elapsedTime)
		ProcessTruckExhaust(truck, truckInput, wheels, exhausts, steams, elapsedTime)
			
		if (truckInput.isEngineIgnited) then
			truck.elapsedStartTime = truck.elapsedStartTime + elapsedTime				
		else			
			truck.elapsedStartTime = 0			
		end
		SpawnWheelsParticles(truck, truckInput.linVelVector, wheels, elapsedTime)		
		SpawnWaveParticles(truck, truckInput, wave, elapsedTime)
		
		C_SetParticlesBrand("snow")
		local nep = GetNumEmitParticles(0.0005, 0.1, elapsedTime)
		for n = 1, nep do
			local upperpoint = m3AddMake(exhausts[1].exhaustOrg, m3MakeVec(20-math.random()*40, 15, 20-math.random()*40))
			local dir = m3MakeVec(2*math.random(), -6, 2*math.random())
			--logVectorDescription("upper", upperpoint)
			--logVectorDescription("direction", dir)
			C_AddParticle(upperpoint, dir)
		end
	end
end
-----------------------------------------------------------------------
--EOF
